// Copyright (C) 2003 Delyan Kalchev
// It's free so do whatever you want!
// Port to PSP by Fanjita Sep 2005
// Port to PSP v2.01 by PSP3D.com Dec 2005
// Port to PSP v2.50, 2.60 by Fanjita Jan 2006

#include <pspdisplay.h>

#include "utils.c"

#define FIELD_WIDTH 12
#define FIELD_HEIGHT 22
#define LEFT TRUE
#define RIGHT FALSE
#define DoLeftM(NewMatrix, OldMatrix, x, y) (NewMatrix[y][3-x]=OldMatrix[x][y])
#define DoRightM(NewMatrix, OldMatrix, x, y) (NewMatrix[3-y][x]=OldMatrix[x][y])
#define IsZeroRowM(Matrix, y) (!Matrix[0][y]&&!Matrix[1][y]&&!Matrix[2][y]&&!Matrix[3][y])
#define IsZeroColM(Matrix, x) (!Matrix[x][0]&&!Matrix[x][1]&&!Matrix[x][2]&&!Matrix[x][3])
#define IsZeroM(Matrix) (IsZeroRowM(Matrix,0)&&IsZeroRowM(Matrix,1)&&IsZeroRowM(Matrix,2)&&IsZeroRowM(Matrix,3))

typedef BOOL Block;

Block Field[FIELD_WIDTH][FIELD_HEIGHT];
Block Matrix[4][4]; //Matrix[X][Y]
Block Next[4][4];
int X, Y;
int m_side;
Block OldField[FIELD_WIDTH][FIELD_HEIGHT];
BOOL Paused;
RECT rect;
int MatrixColour;
int NextColour;
int glines;

#define INDEX_WHITE   1 
#define INDEX_MAGENTA 2
#define INDEX_RED     3
#define INDEX_BLUE    4
#define INDEX_GREEN   5
#define INDEX_YELLOW  6
#define INDEX_CYAN    7

unsigned long colours[9] =
{
  0x00000000L,
  0x00FFFFFFL,
  0x00FF00FFL,
  0x000000FFL,
  0x00FF0000L,
  0x0000FF00L,
  0x0000FFFFL,
  0x00FFFF00L,
  0x00FFFF88L
};

unsigned long score;

RECT fieldrect;
SIZE fieldsize;
RECT nextrect;

int  stHeld;
int  sqHeld;
int  xHeld;
int  oHeld;

void CopyRowM(Block [4][4], int, Block [4][4], int);
void CopyColM(Block [4][4], int, Block [4][4], int);
void Copy();
void NormalizeM(Block [4][4]);
void DeleteRowsM(Block [4][4], int, int);
void DeleteColsM(Block [4][4], int, int);
int WidthM();

void DrawBlock(int, int, BOOL);
void DrawM();

void DelRows();
BOOL OverlapM(Block [4][4], int, int);
void DoDie();
void Incorp(Block [FIELD_WIDTH][FIELD_HEIGHT]);
void DoRect(RECT *, int, int);
void ReSetM();
void Full();
void SetBitsM(Block[4][4], int colourindex,
        Block, Block, Block, Block,
	 		  Block, Block, Block, Block,
			  Block, Block, Block, Block,
			  Block, Block, Block, Block);
void RotateM(BOOL, BOOL, int);
void New();
void Fall();
void MoveLeft();
void MoveRight();
void MoveDown();
void Draw();

void CopyRowM(Block ToMatrix[4][4], int y0, Block FromMatrix[4][4], int y)
{	
	ToMatrix[0][y0]=FromMatrix[0][y];
	ToMatrix[1][y0]=FromMatrix[1][y];
	ToMatrix[2][y0]=FromMatrix[2][y];
	ToMatrix[3][y0]=FromMatrix[3][y];
}

void CopyColM(Block ToMatrix[4][4], int x0, Block FromMatrix[4][4], int x)
{
	ToMatrix[x0][0]=FromMatrix[x][0];
	ToMatrix[x0][1]=FromMatrix[x][1];
	ToMatrix[x0][2]=FromMatrix[x][2];
	ToMatrix[x0][3]=FromMatrix[x][3];
}

void SetBitsM(Block xoMatrix[4][4], int xicolourindex,
              Block a, Block b, Block c, Block d,
    		   	  Block e, Block f, Block g, Block h,
       			  Block i, Block j, Block k, Block l,
       			  Block m, Block n, Block o, Block p)
{
	xoMatrix[0][0] = a? xicolourindex : 0; xoMatrix[1][0] = b? xicolourindex : 0;
  xoMatrix[2][0] = c? xicolourindex : 0; xoMatrix[3][0] = d? xicolourindex : 0;
	xoMatrix[0][1] = e? xicolourindex : 0; xoMatrix[1][1] = f? xicolourindex : 0;
  xoMatrix[2][1] = g? xicolourindex : 0; xoMatrix[3][1] = h? xicolourindex : 0;
	xoMatrix[0][2] = i? xicolourindex : 0; xoMatrix[1][2] = j? xicolourindex : 0;
  xoMatrix[2][2] = k? xicolourindex : 0; xoMatrix[3][2] = l? xicolourindex : 0;
	xoMatrix[0][3] = m? xicolourindex : 0; xoMatrix[1][3] = n? xicolourindex : 0;
  xoMatrix[2][3] = o? xicolourindex : 0; xoMatrix[3][3] = p? xicolourindex : 0;
}

void NormalizeM(Block NewMatrix[4][4])
{
	int i=0;
	while (IsZeroRowM(NewMatrix, i) && i++<4);
	if (i<5 && i>=1) DeleteRowsM(NewMatrix, 0, i-1);
	i=0;
	while (IsZeroColM(NewMatrix, i) && i++<4);
	if (i<5 && i>=1) DeleteColsM(NewMatrix, 0, i-1);
}

void DeleteRowsM(Block NewMatriX[4][4], int from, int to)
{
	if ((from > to) || (to-from>3)) return;
	Block NewMatrix[4][4] = { 0, 0, 0, 0,
							  0, 0, 0, 0,
							  0, 0, 0, 0,
							  0, 0, 0, 0 };
	int y=0;
	const int dif=(to-from)+1;
	for (; y < from; y++) CopyRowM(NewMatrix, y, NewMatriX, y);
	for (y=to+1; y < 4; y++) CopyRowM(NewMatrix, y-dif, NewMatriX, y);
	CopyColM(NewMatriX, 0, NewMatrix, 0);
	CopyColM(NewMatriX, 1, NewMatrix, 1);
	CopyColM(NewMatriX, 2, NewMatrix, 2);
	CopyColM(NewMatriX, 3, NewMatrix, 3);
}

void DeleteColsM(Block NewMatriX[4][4], int from, int to)
{
	if ((from > to) || (to-from>3)) return;
	Block NewMatrix[4][4] = { 0, 0, 0, 0,
							  0, 0, 0, 0,
							  0, 0, 0, 0,
							  0, 0, 0, 0 };
	int x=0;
	const int dif=(to-from)+1;
	for (; x < from; x++) CopyColM(NewMatrix, x, NewMatriX, x);
	for (x=to+1; x < 4; x++) CopyColM(NewMatrix, x-dif, NewMatriX, x);
	CopyColM(NewMatriX, 0, NewMatrix, 0);
	CopyColM(NewMatriX, 1, NewMatrix, 1);
	CopyColM(NewMatriX, 2, NewMatrix, 2);
	CopyColM(NewMatriX, 3, NewMatrix, 3);
}

void RotateM(BOOL z, BOOL rdraw, int xicolour)
{
	if (rdraw) Copy();
	Block NewMatrix[4][4];
	int x, y;
	for (y=0; y<4; y++) {
		for (x=0; x<4; x++) {
			z==LEFT?DoLeftM(NewMatrix, Matrix, x, y) :
					DoRightM(NewMatrix, Matrix, x, y);
		}
	}
	NormalizeM(NewMatrix);
	if (OverlapM(NewMatrix, X, Y) && rdraw) return;
	SetBitsM(Matrix, xicolour,
           NewMatrix[0][0],NewMatrix[1][0],NewMatrix[2][0],NewMatrix[3][0],
           NewMatrix[0][1],NewMatrix[1][1],NewMatrix[2][1],NewMatrix[3][1],
           NewMatrix[0][2],NewMatrix[1][2],NewMatrix[2][2],NewMatrix[3][2],
           NewMatrix[0][3],NewMatrix[1][3],NewMatrix[2][3],NewMatrix[3][3]);
}

void ReSetM()
{
	int i=rand()%4;

  memcpy(Matrix, Next, sizeof(Matrix));
  MatrixColour = NextColour;
  
	switch (rand()%7) {
	case 0:
    NextColour = INDEX_BLUE;
		SetBitsM(Next, NextColour,
         0, 1, 0, 0,
				 0, 1, 0, 0,
				 0, 1, 0, 0,
				 0, 1, 0, 0);
		break;
	case 1:
    NextColour = INDEX_RED;
		SetBitsM(Next, NextColour,
         1, 1, 0, 0,
				 1, 1, 0, 0,
				 0, 0, 0, 0,
				 0, 0, 0, 0);
		break;
	case 2:
    NextColour = 2;
		SetBitsM(Next, NextColour,
         0, 1, 0, 0,
				 1, 1, 1, 0,
				 0, 0, 0, 0,
				 0, 0, 0, 0);
		break;
	case 3:
    NextColour = 8;
		SetBitsM(Next, NextColour,
         1, 1, 0, 0,
				 0, 1, 1, 0,
				 0, 0, 0, 0,
				 0, 0, 0, 0);
		break;
	case 4:
    NextColour = 5;
		SetBitsM(Next, NextColour,
         0, 1, 1, 0,
				 1, 1, 0, 0,
				 0, 0, 0, 0,
				 0, 0, 0, 0);
		break;
	case 5:
    NextColour = 6;
		SetBitsM(Next, NextColour,
         1, 0, 0, 0,
				 1, 1, 1, 0,
				 0, 0, 0, 0,
				 0, 0, 0, 0);
		break;
	case 6:
    NextColour = 7;
		SetBitsM(Next, NextColour,
         0, 0, 1, 0,
				 1, 1, 1, 0,
				 0, 0, 0, 0,
				 0, 0, 0, 0);
	}
//	for (; i > 0; i--) RotateM(LEFT, FALSE, NextColour);
	X=(int)(FIELD_WIDTH-WidthM())/2;
	Y=FIELD_HEIGHT-1;
	if (OverlapM(Matrix, X, Y))
  {
    DoDie();
  }
}

void Full()
{
	X=Y=0;
	SetBitsM(Matrix, 0,
       0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0);
	SetBitsM(Next, 0,
       0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0);
	int x, y=0;
	for (; y<FIELD_HEIGHT; y++)
		for(x=0; x<FIELD_WIDTH; x++)
			Field[x][y]=INDEX_WHITE;
}

void New()
{
	X=Y=0;
	SetBitsM(Matrix, 0,
       0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0);
	SetBitsM(Next, 0,
       0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0,
			 0, 0, 0, 0);
	int x, y=0;
	for (; y<FIELD_HEIGHT; y++)
		for(x=0; x<FIELD_WIDTH; x++)
			Field[x][y]=0;
	ReSetM();
  ReSetM();
}

void DrawBlock(int x, int y, unsigned char present)
{
	RECT Rect;
	x=x*m_side;
	y=(FIELD_HEIGHT - (y+1))*m_side;

	Rect.left = x + fieldrect.left;
	Rect.top = y + fieldrect.top;
	Rect.right = x+m_side + fieldrect.left;
	Rect.bottom = y+m_side + fieldrect.top;

  if (present)
  {
  	FillRect(&Rect, TEXTCOLOUR);
  
  	Rect.left += 1;
  	Rect.top += 1;
  	Rect.right -= 1;
  	Rect.bottom -= 1;
  	FillRect(&Rect, colours[present]);
  }
  else
  {
  	FillRect(&Rect, 0);
  }
}

void DrawBlockNext(int x, int y, unsigned char present)
{
	RECT Rect;
	x=x*m_side;
	y=y*m_side;


	Rect.left = x + nextrect.left;
	Rect.top = y + nextrect.top;
	Rect.right = x+m_side + nextrect.left;
	Rect.bottom = y+m_side + nextrect.top;

  if (present)
  {
  	FillRect(&Rect, TEXTCOLOUR);
  
  	Rect.left += 1;
  	Rect.top += 1;
  	Rect.right -= 1;
  	Rect.bottom -= 1;
  	FillRect(&Rect, colours[present]);
  }
  else
  {
  	FillRect(&Rect, 0);
  }
}

void Draw()
{
  RECT  lfillrect;
	int x, y=0;

	for (; y<FIELD_HEIGHT; y++)
  {
		for(x=0; x<FIELD_WIDTH; x++)
    {
			DrawBlock(x, y, Field[x][y]);
    }
  }

	if (!IsZeroM(Matrix)) DrawM();

  lfillrect.left   = fieldrect.left - m_side;
  lfillrect.right  = fieldrect.right + m_side;
  lfillrect.top    = fieldrect.bottom;
  lfillrect.bottom = fieldrect.bottom + m_side;
  FillRect(&lfillrect, BLUE);

  lfillrect.left   = fieldrect.left - m_side;
  lfillrect.right  = fieldrect.left;
  lfillrect.top    = fieldrect.top;
  lfillrect.bottom = fieldrect.bottom;
  FillRect(&lfillrect, BLUE);

  lfillrect.left   = fieldrect.right;
  lfillrect.right  = fieldrect.right+m_side;
  lfillrect.top    = fieldrect.top;
  lfillrect.bottom = fieldrect.bottom;
  FillRect(&lfillrect, BLUE);

}

void DrawM()
{
	char x, y=0;
	for (; y<4; y++)
		for(x=0; x<4; x++)
			if (Matrix[x][y]) DrawBlock(X+x, Y-y, Matrix[x][y]);
}

void Copy()
{
	int i=0, j;
	for (; i<FIELD_HEIGHT; i++)
		for (j=0; j<FIELD_WIDTH; j++)
			OldField[j][i]=Field[j][i];
	Incorp(OldField);
}

void DoRect(RECT *Rect, int x, int y)
{
	x=x*m_side;
	y=(FIELD_HEIGHT - (y+1))*m_side;
	Rect->left = x-1;
	Rect->top = y-1;
	Rect->right = x+m_side+1;
	Rect->bottom = y+m_side+1;
}

void Fall()
{
	if (IsZeroM(Matrix)) return;
	Copy();
	for (; !OverlapM(Matrix, X, Y); Y--); Y++;
	Incorp(Field);
	DelRows();
	Copy();
	ReSetM();
}

void Incorp(Block Fld[FIELD_WIDTH][FIELD_HEIGHT])
{
	if (IsZeroM(Matrix)) return;
	int i=0, j;
	for (; i<4; i++)
		for (j=0; j<4; j++)
			if (Matrix[j][i] && Y-i>=0)
				Fld[X+j][Y-i] = Matrix[j][i];
}

void DelRows()
{
	int i, j, k, count=0;
	Copy();
	for (i=0; i<FIELD_HEIGHT; i++)
  {
		for (j=0; j<FIELD_WIDTH && Field[j][i]; j++)                               
    {
    }

  	if (j==FIELD_WIDTH)
    {
      count ++;
  		for (j=i; j<FIELD_HEIGHT-1; j++)
  			for (k=0; k<FIELD_WIDTH; k++) Field[k][j]=Field[k][j+1];
  		for (k=0; k<FIELD_WIDTH; k++) Field[k][j]=0;
  		i--;
  	}
  }

  switch (count)
  {
    case 0:
      break;
    case 1:
      score += 10;
      break;
    case 2:
      score += 50;
      break;
    case 3:
      score += 100;
      break;
    case 4:
      score += 1000;
      break;
    default:
      break;
  }

  glines += count;
}

int WidthM()
{
	int max=0, i, j=0;
	for (;j<3;j++) {
		for (i=4; i>0 && !Matrix[i-1][j]; i--);
		max = i>max?i:max;
	}
	return max;
}

void MoveLeft()
{
	Copy();
	if (!OverlapM(Matrix, X-1, Y)) X--;
}

void MoveRight()
{
	Copy();
	if (!OverlapM(Matrix, X+1, Y)) X++;
}

void MoveDown()
{
	Copy();
	if (OverlapM(Matrix, X, Y-1)) { Fall(); return;}
	Y--;
}

void DoDie()
{
	Full();
  Draw();
  changeBuffer();
  Full();
  Draw();

  Sleep(2000);

  changeBuffer();

  New();
  Draw();
  changeBuffer();
  Draw();
  changeBuffer();

  score = 0;
  glines = 0;
}

BOOL OverlapM(Block NewMatrix[4][4], int x, int y)
{
	if (IsZeroM(NewMatrix)) return FALSE;
	int i, j;
	for (i=0; i<4; i++)
		for (j=0; j<4; j++) {
			if (NewMatrix[j][i]) {
				if (x+j<0 || x+j>=FIELD_WIDTH || y-i<0) return TRUE;
				if (Field[x+j][y-i]) return TRUE;
			}
		}
	return FALSE;
}

void ProcessKeys(unsigned long xikeys)
{
  if (!Paused)
  {
    if (xikeys & PSP_CTRL_LEFT)
    {
      MoveLeft();
    }
    if (xikeys & PSP_CTRL_RIGHT)
    {
      MoveRight();
    }
    if (xikeys & PSP_CTRL_DOWN)
    {
      MoveDown();
    }

    if (xikeys & PSP_CTRL_CROSS)
    {
      if (!xHeld)
      {
        RotateM(RIGHT, TRUE, MatrixColour);
        xHeld = 1;
      }
    }
    else
    {
      xHeld = 0;
    }

    if (xikeys & PSP_CTRL_CIRCLE)
    {
      if (!oHeld)
      {
        RotateM(LEFT, TRUE, MatrixColour);
        oHeld = 1;
      }
    }
    else
    {
      oHeld = 0;
    }

    if (xikeys & PSP_CTRL_SQUARE)
    {
      if (!sqHeld)
      {
        Fall();
        sqHeld = 1;
      }
    }
    else
    {
      sqHeld = 0;
    }
  }

  if (xikeys & PSP_CTRL_START)
  {
    if (!stHeld)
    {
      if (!Paused)
      {
        Paused = TRUE;
      }
      else if (Paused)
      {
        Paused = FALSE;
      }
      stHeld = 1;
    }
  }
  else
  {
    stHeld = 0;
  }
}

/*****************************************************************************/
/* Hacky sleep routine                                                       */
/*                                                                           */
/* Better to convert this to a wait-for-vsync or something                   */
/*****************************************************************************/
int Sleep(unsigned int xicount)
{
  int ldummy = 2;
  unsigned int i;
  unsigned int j;

  for (i=0; i < xicount; i++)
  {
    for (j=0; j < 65000; j++)
    {
      ldummy = ldummy * ldummy;
    }
  }

  return(ldummy);
}

void ShowNext()
{
  int x, y;
  for (x=0; x<4; x++)
  {
    for (y=0; y<4; y++)
    {
      DrawBlockNext(x,y,Next[x][y]);
    }
  }
}

void PrintInstrucs()
{
  Print(3,0,GREEN,"Controls:");
  Print(3,1,GREEN,"---------");
  Print(0,3,GREEN,"O  : Rotate");
  Print(0,4,GREEN,"X  : Rotate");
  Print(0,5,GREEN,"[] : Drop");
  Print(0,6,GREEN,"<- : Move Left");
  Print(0,7,GREEN,"-> : Move Right");
  Print(0,8,GREEN,"| ");
  Print(0,9,GREEN,"v  : Move Down");

  if (Paused)
  {
    Print(3,12,RED,"PAUSED");
  }
  else
  {
    Print(3,12,0,"PAUSED");
  }

  Print(43, 0,0x00888888,"PSP Tetris v2.0");
  Print(43, 2,0x00888888," 11 Jan 2006");
  Print(43, 4,0x00888888,"by Fanjita");
  Print(43, 5,0x00888888,"www.fanjita.org");

  Print(43, 7,0x00888888,"Includes code");
  Print(43, 8,0x00888888,"from:        ");
  Print(43, 9,0x00888888," Groepaz     ");
  Print(43,10,0x00888888," PSPSDK      ");
  Print(43,11,0x00888888," Delyan Kalchev");
  Print(43,12,0x00888888," Ditlew");
  Print(43,13,0x00888888," Edison Carter");
  Print(43,14,0x00888888," Bitmap1");

  Print(43,17,0x00FFFFFF,"Score:");
  WriteDecimal(46, 18, score);
  Print(43,19,0x00FFFFFF,"Lines:");
  WriteDecimal(46, 20, glines);

  Print(43,22,0x00FFFFFF,"Next:");
  ShowNext();
}

void RectCpy(RECT *ldest, RECT *lsrc)
{
  ldest->left   = lsrc->left;
  ldest->right  = lsrc->right;
  ldest->top    = lsrc->top;
  ldest->bottom = lsrc->bottom;
}

void fadeIn()
{
  int i;

  for (i = 0; i < 16; i+=3)
  {
    Fillvram(0x00111111 * i);
    sceDisplayWaitVblankStart();
    changeBuffer();
  }
}

void fadeOut()
{
  int i;

  for (i = 15; i >= 0; i--)
  {
    Fillvram(0x00111111 * i);
    sceDisplayWaitVblankStart();
    changeBuffer();
  }
}

void Main(void)
{
	RECT deskrect, wndrect, clientrect;
	SIZE desksize, wndsize;
	POINT pt;
  int i;

//  flashscreen(0, 0x00ff0000L);

  deskrect.left = 0;
  deskrect.right = 480;
  deskrect.bottom = 272;
  deskrect.top = 0;

  score = 0;

	desksize.cy = deskrect.bottom - deskrect.top;
	desksize.cx = deskrect.right - deskrect.left;
	m_side = (int)((desksize.cy - 10) / (FIELD_HEIGHT+2));

  RectCpy(&wndrect,&deskrect);
  RectCpy(&clientrect,&deskrect);
	pt.x = clientrect.left;
	pt.y = clientrect.top;
	clientrect.left = pt.x;
	clientrect.top = pt.y;
	pt.x = clientrect.right;
	pt.y = clientrect.bottom;
	clientrect.right = pt.x;
	clientrect.bottom = pt.y;
	wndsize.cx = (clientrect.left - wndrect.left) + (wndrect.right - clientrect.right) + FIELD_WIDTH * m_side;
	wndsize.cy = (clientrect.top - wndrect.top) + (wndrect.bottom - clientrect.bottom) + FIELD_HEIGHT * m_side;
	Paused=FALSE;

  fieldsize.cx = m_side * FIELD_WIDTH;
  fieldsize.cy = m_side * FIELD_HEIGHT;

  fieldrect.left = (clientrect.left + clientrect.right) / 2    // centre
                   - (fieldsize.cx / 2);
  fieldrect.right = fieldrect.left + fieldsize.cx;
  fieldrect.top  = 0;
  fieldrect.bottom = fieldrect.top + fieldsize.cy;

  nextrect.left = 43 * 8;
  nextrect.top  = 23 * 8;
  nextrect.right = nextrect.left + 4 * m_side;
  nextrect.bottom = nextrect.top + 4 * m_side;

  NextColour = 0;
  MatrixColour = 0;

  /***************************************************************************/
  /* Set up screen and controller                                            */
  /***************************************************************************/
  initScreenAndCtrl();

  fadeIn();

  fadeOut();

  stHeld = 0;
  sqHeld = 0;
  xHeld  = 0;
  oHeld  = 0;

  glines = 0;

  Full();
  Draw();
  changeBuffer();

  New();
  changeBuffer();

  score = 0;

  while(1)
  {
    PrintInstrucs();

    i ++;
    if ((i % (glines < 126? (64 - (glines / 2)) : 1)) == 0)
    {
      if (!Paused)
      {
        MoveDown();
      }
    }

    gpaddata.Buttons = 0;

    sceCtrlReadBufferPositive(&gpaddata,1);

    WriteDwordDummy(0,22,gpaddata.Buttons);

    if ((i % 6) == 0)
    {
      ProcessKeys(gpaddata.Buttons);
    }
                                
    Draw();

    changeBuffer();
  }
}
